unit libYubikey;

(* yubikey.h --- Prototypes for low-level Yubikey OTP functions.
 *
 * Written by Simon Josefsson <simon@josefsson.org>.
 * Copyright (c) 2006, 2007, 2008, 2009 Yubico AB
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *    * Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
(* Delphi version by Pepak
 * v0.01, based on libyubikey v??
 *)

{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

interface

const
  YUBIKEY_BLOCK_SIZE = 16;
  YUBIKEY_KEY_SIZE   = 16;  // bytes
  YUBIKEY_UID_SIZE   =  6;

const
  YUBIKEY_MODHEX_MAP = 'cbdefghijklnrtuv';

const
  YUBIKEY_CRC_OK_RESIDUE = $f0b8;

type
  PYubikeyUID = ^TYubikeyUID;
  TYubikeyUID = packed array [0..YUBIKEY_UID_SIZE-1] of byte;

  PYubikeyToken = ^TYubikeyToken;
  TYubikeyToken = packed record
    case Boolean of
      True:
        (
          // Unique (secret) ID.
          uid: TYubikeyUID;
          // Session counter (incremented by 1 at each startup).  High bit
          // indicates whether caps-lock triggered the token.
          ctr: word;
          // Timestamp incremented by approx 8Hz (low part).
          tstpl: word;
          // Timestamp (high part).
          tstph: byte;
          // Number of times used within session + activation flags.
          use: byte;
          // Pseudo-random value.
          rnd: word;
          // CRC16 value of all fields.
          crc: word;
        );
      False:
        (
          bytes: array[0..YUBIKEY_BLOCK_SIZE-1] of byte
        );
    end;

  PYubikeyOTP = ^TYubikeyOTP;
  TYubikeyOTP = packed array [0..YUBIKEY_BLOCK_SIZE*2-1] of byte;

  PYubikeyKey = ^TYubikeyKey;
  TYubikeyKey = packed array [0..YUBIKEY_KEY_SIZE-1] of byte;

(* Decrypt TOKEN using KEY and store output in OUT structure.  Note
   that there is no error checking whether the output data is valid or
   not, use yubikey_check_* for that. *)
procedure YubikeyParse(OTP: PYubikeyOTP; Key: PYubikeyKey; Parsed: PYubikeyToken);

(* ModHex encode input string SRC of length SRCSIZE and put the zero
   terminated output string in DST.  The size of the output string DST
   must be at least 2*SRCSIZE+1.  The output string is always
   2*SRCSIZE large plus the terminating zero.  *)
procedure YubikeyModhexEncode(Dest, Src: PChar; SrcSize: integer);

(* ModHex decode input string SRC of length DSTSIZE/2 into output
   string DST.  The output string DST is always DSTSIZE/2 large plus
   the terminating zero.  *)
procedure YubikeyModhexDecode(Dest, Src: PChar; DestSize: integer);

(* Hex encode/decode data, same interface as modhex functions. *)
procedure YubikeyHexEncode(Dest, Src: PChar; SrcSize: integer);
procedure YubikeyHexDecode(Dest, Src: PChar; DestSize: integer);

function YubikeyCRC16(Buffer: PChar; BufSize: integer): word;

(* AES-decrypt one 16-byte block STATE using the 128-bit KEY, leaving
   the decrypted output in the STATE buffer. *)
procedure YubikeyAesDecrypt(State: PYubikeyToken; Key: PYubikeyKey);

implementation

uses
  uAnsiStrings;
  
procedure YubikeyEncode(Dest, Src: PChar; SrcSize: integer; Map: PChar);
begin
  while SrcSize > 0 do
    begin
    Dest^ := Map[(Byte(Src^) shr 4) and $f];
    Inc(Dest);
    Dest^ := Map[Byte(Src^) and $f];
    Inc(Dest);
    Inc(Src);
    Dec(SrcSize);
    end;
end;

procedure YubikeyDecode(Dest, Src: PChar; DestSize: integer; Map: PChar);
var P: PChar;
    C: Char;
    B: Byte;
    FirstHalf: boolean;
begin
  FirstHalf := True;
  while (Src^ <> #0) and (DestSize > 0) do
    begin
    C := LowCase(Src^);
    P := StrScan(Map, C);
    if P = nil then
      B := 0
    else
      B := Longword(P) - Longword(Map);
    if FirstHalf then
      Dest^ := Char(B)
    else
      begin
      Dest^ := Char((Byte(Dest^) shl 4) or B);
      Inc(Dest);
      Dec(DestSize);
      end;
    FirstHalf := not FirstHalf;
    Inc(Src);
    end;
  while DestSize > 0 do
    begin
    Dest^ := #0;
    Inc(Dest);
    Dec(DestSize);
    end;
end;

procedure YubikeyModhexEncode(Dest, Src: PChar; SrcSize: integer);
begin
  YubikeyEncode(Dest, Src, SrcSize, YUBIKEY_MODHEX_MAP);
end;

procedure YubikeyModhexDecode(Dest, Src: PChar; DestSize: integer);
begin
  YubikeyDecode(Dest, Src, DestSize, YUBIKEY_MODHEX_MAP);
end;

const
  YUBIKEY_HEX_MAP = '0123456789abcdef';

procedure YubikeyHexEncode(Dest, Src: PChar; SrcSize: integer);
begin
  YubikeyEncode(Dest, Src, SrcSize, YUBIKEY_HEX_MAP);
end;

procedure YubikeyHexDecode(Dest, Src: PChar; DestSize: integer);
begin
  YubikeyDecode(Dest, Src, DestSize, YUBIKEY_HEX_MAP);
end;

function YubikeyCRC16(Buffer: PChar; BufSize: integer): word;
var i, j: integer;
begin
  Result := $ffff;
  while BufSize > 0 do
    begin
    Result := Result xor Byte(Buffer^);
    for i := 0 to 7 do
      begin
      j := Result and $1;
      Result := Result shr 1;
      if j > 0 then
        Result := Result xor $8408;
      end;
    Inc(Buffer);
    Dec(BufSize);
    end;
end;

procedure YubikeyParse(OTP: PYubikeyOTP; Key: PYubikeyKey; Parsed: PYubikeyToken);
begin
  FillChar(Parsed^, Sizeof(Parsed^), 0);
  YubikeyModhexDecode(Pointer(Parsed), Pointer(OTP), Sizeof(Parsed^));
  YubikeyAesDecrypt(Parsed, Key);
end;

procedure YubikeyAesDecrypt(State: PYubikeyToken; Key: PYubikeyKey);
const NUMBER_OF_ROUNDS = 10;
const RC: array[0..NUMBER_OF_ROUNDS-1] of byte = ( $1, $2, $4, $8, $10, $20, $40, $80, $1B, $36 );
const rijndael_sbox: array[byte] of byte = (
          $63, $7C, $77, $7B, $F2, $6B, $6F, $C5,
          $30, $01, $67, $2B, $FE, $D7, $AB, $76,
          $CA, $82, $C9, $7D, $FA, $59, $47, $F0,
          $AD, $D4, $A2, $AF, $9C, $A4, $72, $C0,
          $B7, $FD, $93, $26, $36, $3F, $F7, $CC,
          $34, $A5, $E5, $F1, $71, $D8, $31, $15,
          $04, $C7, $23, $C3, $18, $96, $05, $9A,
          $07, $12, $80, $E2, $EB, $27, $B2, $75,
          $09, $83, $2C, $1A, $1B, $6E, $5A, $A0,
          $52, $3B, $D6, $B3, $29, $E3, $2F, $84,
          $53, $D1, $00, $ED, $20, $FC, $B1, $5B,
          $6A, $CB, $BE, $39, $4A, $4C, $58, $CF,
          $D0, $EF, $AA, $FB, $43, $4D, $33, $85,
          $45, $F9, $02, $7F, $50, $3C, $9F, $A8,
          $51, $A3, $40, $8F, $92, $9D, $38, $F5,
          $BC, $B6, $DA, $21, $10, $FF, $F3, $D2,
          $CD, $0C, $13, $EC, $5F, $97, $44, $17,
          $C4, $A7, $7E, $3D, $64, $5D, $19, $73,
          $60, $81, $4F, $DC, $22, $2A, $90, $88,
          $46, $EE, $B8, $14, $DE, $5E, $0B, $DB,
          $E0, $32, $3A, $0A, $49, $06, $24, $5C,
          $C2, $D3, $AC, $62, $91, $95, $E4, $79,
          $E7, $C8, $37, $6D, $8D, $D5, $4E, $A9,
          $6C, $56, $F4, $EA, $65, $7A, $AE, $08,
          $BA, $78, $25, $2E, $1C, $A6, $B4, $C6,
          $E8, $DD, $74, $1F, $4B, $BD, $8B, $8A,
          $70, $3E, $B5, $66, $48, $03, $F6, $0E,
          $61, $35, $57, $B9, $86, $C1, $1D, $9E,
          $E1, $F8, $98, $11, $69, $D9, $8E, $94,
          $9B, $1E, $87, $E9, $CE, $55, $28, $DF,
          $8C, $A1, $89, $0D, $BF, $E6, $42, $68,
          $41, $99, $2D, $0F, $B0, $54, $BB, $16
        );
const rijndael_inv_sbox: array[byte] of byte = (
          $52, $09, $6A, $D5, $30, $36, $A5, $38,
          $BF, $40, $A3, $9E, $81, $F3, $D7, $FB,
          $7C, $E3, $39, $82, $9B, $2F, $FF, $87,
          $34, $8E, $43, $44, $C4, $DE, $E9, $CB,
          $54, $7B, $94, $32, $A6, $C2, $23, $3D,
          $EE, $4C, $95, $0B, $42, $FA, $C3, $4E,
          $08, $2E, $A1, $66, $28, $D9, $24, $B2,
          $76, $5B, $A2, $49, $6D, $8B, $D1, $25,
          $72, $F8, $F6, $64, $86, $68, $98, $16,
          $D4, $A4, $5C, $CC, $5D, $65, $B6, $92,
          $6C, $70, $48, $50, $FD, $ED, $B9, $DA,
          $5E, $15, $46, $57, $A7, $8D, $9D, $84,
          $90, $D8, $AB, $00, $8C, $BC, $D3, $0A,
          $F7, $E4, $58, $05, $B8, $B3, $45, $06,
          $D0, $2C, $1E, $8F, $CA, $3F, $0F, $02,
          $C1, $AF, $BD, $03, $01, $13, $8A, $6B,
          $3A, $91, $11, $41, $4F, $67, $DC, $EA,
          $97, $F2, $CF, $CE, $F0, $B4, $E6, $73,
          $96, $AC, $74, $22, $E7, $AD, $35, $85,
          $E2, $F9, $37, $E8, $1C, $75, $DF, $6E,
          $47, $F1, $1A, $71, $1D, $29, $C5, $89,
          $6F, $B7, $62, $0E, $AA, $18, $BE, $1B,
          $FC, $56, $3E, $4B, $C6, $D2, $79, $20,
          $9A, $DB, $C0, $FE, $78, $CD, $5A, $F4,
          $1F, $DD, $A8, $33, $88, $07, $C7, $31,
          $B1, $12, $10, $59, $27, $80, $EC, $5F,
          $60, $51, $7F, $A9, $19, $B5, $4A, $0D,
          $2D, $E5, $7A, $9F, $93, $C9, $9C, $EF,
          $A0, $E0, $3B, $4D, $AE, $2A, $F5, $B0,
          $C8, $EB, $BB, $3C, $83, $53, $99, $61,
          $17, $2B, $04, $7E, $BA, $77, $D6, $26,
          $E1, $69, $14, $63, $55, $21, $0C, $7D
        );

  function XTime(B: Byte): Byte; assembler;
  asm
        TEST AL, $80
        JE @NoXor
        SHL AL, 1
        XOR AL, $1b
        JMP @End
  @NoXor:
        SHL AL, 1
  @End:
  end;

var i, j: byte;
    a02x, a13x, a02xx, a13xx, k1, k2: byte;
    round_key: array[0..15] of byte;
begin
  // Key preparation
  Move(Key^, round_key, sizeof(round_key));
  for i := 0 to Pred(NUMBER_OF_ROUNDS) do
    begin
    round_key[0] := round_key[0] xor RC[i];
    round_key[0] := round_key[0] xor rijndael_sbox[round_key[13]];
    round_key[1] := round_key[1] xor rijndael_sbox[round_key[14]];
    round_key[2] := round_key[2] xor rijndael_sbox[round_key[15]];
    round_key[3] := round_key[3] xor rijndael_sbox[round_key[12]];
    for j := 4 to 15 do
      round_key[j] := round_key[j] xor round_key[j - 4];
    end;
  // Decryption
  for i := 0 to 15 do
    State^.bytes[i] := State^.bytes[i] xor round_key[i];
  for i := 1 to NUMBER_OF_ROUNDS do
    begin
    // inv_byte_sub_shift_row();

    // First row: 0 shift, 0 4 8 12
    State^.bytes[0] := rijndael_inv_sbox[State^.bytes[0]];
    State^.bytes[4] := rijndael_inv_sbox[State^.bytes[4]];
    State^.bytes[8] := rijndael_inv_sbox[State^.bytes[8]];
    State^.bytes[12] := rijndael_inv_sbox[State^.bytes[12]];
    // Second row: -1 shift, 1 5 9 13
    j := State^.bytes[13];
    State^.bytes[13] := rijndael_inv_sbox[State^.bytes[9]];
    State^.bytes[9] := rijndael_inv_sbox[State^.bytes[5]];
    State^.bytes[5] := rijndael_inv_sbox[State^.bytes[1]];
    State^.bytes[1] := rijndael_inv_sbox[j];
    // Third row: -2 shift, 2 6 10 14
    j := State^.bytes[2];
    State^.bytes[2] := rijndael_inv_sbox[State^.bytes[10]];
    State^.bytes[10] := rijndael_inv_sbox[j];
    j := State^.bytes[6];
    State^.bytes[6] := rijndael_inv_sbox[State^.bytes[14]];
    State^.bytes[14] := rijndael_inv_sbox[j];
    // Fourth row: -3 shift, 3 7 11 15
    j := State^.bytes[3];
    State^.bytes[3] := rijndael_inv_sbox[State^.bytes[7]];
    State^.bytes[7] := rijndael_inv_sbox[State^.bytes[11]];
    State^.bytes[11] := rijndael_inv_sbox[State^.bytes[15]];
    State^.bytes[15] := rijndael_inv_sbox[j];

    // get_inv_round_key(i);
    for j := 15 downto 4 do
      round_key[j] := round_key[j] xor round_key[j - 4];
    round_key[0] := round_key[0] xor (RC[NUMBER_OF_ROUNDS - i] xor rijndael_sbox[round_key[13]]);
    round_key[1] := round_key[1] xor rijndael_sbox[round_key[14]];
    round_key[2] := round_key[2] xor rijndael_sbox[round_key[15]];
    round_key[3] := round_key[3] xor rijndael_sbox[round_key[12]];

    for j := 0 to 15 do
      State^.bytes[j] := State^.bytes[j] xor round_key[j];
    if (i <> NUMBER_OF_ROUNDS) then
      begin
      //inv_mix_column();

      j := 0;
      while j < 16 do
        begin
        k1 := State^.bytes[j] xor State^.bytes[j + 2];
        a02x := xtime (k1);
        k2 := State^.bytes[j + 1] xor State^.bytes[j + 3];
        a13x := xtime (k2);

        k1 := k1 xor (k2 xor xtime (State^.bytes[j + 1] xor State^.bytes[j + 2]));
        k2 := k1;

        a02xx := xtime (a02x);
        a13xx := xtime (a13x);

        k1 := k1 xor (xtime (a02xx xor a13xx) xor a02xx);
        k2 := k2 xor (xtime (a02xx xor a13xx) xor a13xx);

        State^.bytes[j] := State^.bytes[j] xor (k1 xor a02x);
        State^.bytes[j + 1] := State^.bytes[j + 1] xor k2;
        State^.bytes[j + 2] := State^.bytes[j + 2] xor (k1 xor a13x);
        State^.bytes[j + 3] := State^.bytes[j + 3] xor (k2 xor a02x xor a13x);

        Inc(j, 4);
        end;
      end;
    end;
end;

end.

