unit uYubikey;
{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

interface

uses
  libYubikey, Windows;

const
  YUBIKEY_ERROR_BASE = 12345678;
  YUBIKEY_ERROR_MAX = YUBIKEY_ERROR_BASE + 10;
  YUBIKEY_ERROR_INVALID_PARAMETERS = YUBIKEY_ERROR_BASE + 1;
  YUBIKEY_ERROR_UNKNOWN_NAME = YUBIKEY_ERROR_BASE + 2;
  YUBIKEY_ERROR_INVALID_OTP = YUBIKEY_ERROR_BASE + 3;
  YUBIKEY_ERROR_INVALID_DATA = YUBIKEY_ERROR_BASE + 4;

(* Register a new Yubikey to the system.
   Input:
       Name: PChar - identifier of the token (e.g. username)
       UID: PYubikeyUID - (secret) unique identifier of the token
       Key: PYubikeyKet - (secret) AES key of the token
       OtpCounter: integer - if larger than zero, current value of the OTP counter
   Output:
       Result: integer - error code of the operation. 0 means successful
                         registration, above codes mean specific Yubikey
                         errors, other values are Windows' GetLastError codes *)
function PKYubikeyRegisterOTP(Name: PChar; UID: PYubikeyUID; Key: PYubikeyKey; OtpCounter: Longint): Longint; stdcall;

(* Register a new Yubikey to the system. It accepts the same parameters and returns
   the same values as PKYoubikeyRegisterOTP, except that UID and Key are passed as
   strings (hex-encoded). *)
function PKYubikeyRegisterOTPhex(Name, UIDHex, KeyHex: PChar; OtpCounter: Longint): Longint; stdcall;

(* Authenticate a token's OTP:
   Input:
       Name: PChar - identifier of the token (e.g. username)
       OTP: PChar - one-time-password gathered from the Yubikey
   Output:
       Result: integer - error code of the operation. 0 means successful
                         authentication, above codes mean specific Yubikey
                         errors, other values are Windows' GetLastError codes *)
function PKYubikeyValidateOTP(Name, OTP: PChar): Longint; stdcall;

implementation

uses
  {$IFDEF DEBUG}
  uDebug,
  {$ENDIF}
  uAnsiStrings;

const
  YUBIKEY_REG_HKEY = HKEY_LOCAL_MACHINE;
  YUBIKEY_REG_ROOTKEY = 'Software\Pepak\Yubikey\'; // Must end with backslash
  YUBIKEY_REG_UID = 'UID';
  YUBIKEY_REG_KEY = 'Key';
  YUBIKEY_REG_COUNTER = 'Counter';
  YUBIKEY_MUTEX = 'PkYubikey.';
  YUBIKEY_MUTEX_TIME = 1000; // Milliseconds

function PKYubikeyRegisterOTPhex(Name, UIDHex, KeyHex: PChar; OtpCounter: Longint): Longint;
var UID: TYubikeyUID;
    Key: TYubikeyKey;
begin
  YubikeyHexDecode(@UID, UIDHex, Sizeof(UID));
  YubikeyHexDecode(@Key, KeyHex, Sizeof(Key));
  Result := PKYubikeyRegisterOTP(Name, @UID, @Key, OtpCounter);
end;

const YUBIKEY_ROOTKEY_LENGTH = Cardinal(Length(YUBIKEY_REG_ROOTKEY));
      YUBIKEY_MUTEX_LENGTH = Cardinal(Length(YUBIKEY_MUTEX));

function OpenRegistryKey(Name: PChar; Create: boolean; out Mutex: THandle; out OwnMutex: boolean; out Reg: HKey): integer;
var MutexName, RegKeyName: PChar;
    NameLen: Cardinal;
begin
  NameLen := StrLen(Name);
  MutexName := StrAlloc(YUBIKEY_MUTEX_LENGTH + NameLen + 1);
  try
    StrCopy(MutexName, YUBIKEY_MUTEX);
    StrCat(MutexName, Name);
    Reg := 0;
    Mutex := CreateMutex(nil, False, MutexName);
    if Mutex <> 0 then
      begin
      Result := WaitForSingleObject(Mutex, YUBIKEY_MUTEX_TIME);
      if Result = WAIT_OBJECT_0 then
        begin
        OwnMutex := True;
        RegKeyName := StrAlloc(YUBIKEY_ROOTKEY_LENGTH + NameLen + 1);
        try
          StrCopy(RegKeyName, YUBIKEY_REG_ROOTKEY);
          StrCat(RegKeyName, Name);
          if Create then
            Result := RegCreateKeyEx(YUBIKEY_REG_HKEY, RegKeyName, 0, nil, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, nil, Reg, nil)
          else
            Result := RegOpenKeyEx(YUBIKEY_REG_HKEY, RegKeyName, 0, KEY_ALL_ACCESS, Reg);
          if Result = ERROR_SUCCESS then
            Result := 0
          else
            Reg := 0;
        finally
          StrDispose(RegKeyName);
          end;
        end
      else
        OwnMutex := False;
      end
    else
      Result := GetLastError;
  finally
    StrDispose(MutexName);
    end;
end;

function CloseRegistryKey(Mutex: THandle; OwnMutex: boolean; Reg: HKey): integer;
begin
  if Reg <> 0 then
    RegCloseKey(Reg);
  if OwnMutex then
    ReleaseMutex(Mutex);
  if Mutex <> 0 then
    CloseHandle(Mutex);
  Result := 0;
end;

function PKYubikeyRegisterOTP(Name: PChar; UID: PYubikeyUID; Key: PYubikeyKey; OtpCounter: Longint): Longint;
var Mutex: THandle;
    OwnMutex: boolean;
    Reg: HKey;
begin
  if (Name = nil) or (Key = nil) then
    Result := YUBIKEY_ERROR_INVALID_PARAMETERS
  else
    begin
    Result := OpenRegistryKey(Name, True, Mutex, OwnMutex, Reg);
    try
      if Result = 0 then
        if (RegSetValueEx(Reg, YUBIKEY_REG_UID, 0, REG_BINARY, UID, Sizeof(UID^)) = ERROR_SUCCESS) and
           (RegSetValueEx(Reg, YUBIKEY_REG_KEY, 0, REG_BINARY, Key, SizeOf(Key^)) = ERROR_SUCCESS) and
           ((OtpCounter < 0) or (RegSetValueEx(Reg, YUBIKEY_REG_COUNTER, 0, REG_DWORD, @OtpCounter, Sizeof(OtpCounter)) = ERROR_SUCCESS))
        then
          Result := 0
        else
          Result := GetLastError;
    finally
      CloseRegistryKey(Mutex, OwnMutex, Reg);
      end;
    end;
end;

function PKYubikeyValidateOTP(Name, OTP: PChar): Longint;
var Mutex: THandle;
    OwnMutex: boolean;
    Reg: HKey;
    UID: TYubikeyUID;
    Key: TYubikeyKey;
    Token: TYubikeyToken;
    UIDSize, KeySize, RegCounterSize, OtpCounter, RegCounter: DWORD;
    b: boolean;
    i: integer;
begin
  if (Name = nil) or (OTP = nil) or (StrLen(OTP) < 32) then
    Result := YUBIKEY_ERROR_INVALID_PARAMETERS
  else
    begin
    Result := OpenRegistryKey(Name, False, Mutex, OwnMutex, Reg);
    try
      if Result = 0 then
        begin
        UIDSize := Sizeof(UID);
        KeySize := Sizeof(Key);
        if (RegQueryValueEx(Reg, YUBIKEY_REG_UID, nil, nil, @UID, @UIDSize) = ERROR_SUCCESS) and
           (RegQueryValueEx(Reg, YUBIKEY_REG_KEY, nil, nil, @Key, @KeySize) = ERROR_SUCCESS)
        then
          if (UIDSize = Sizeof(UID)) and (KeySize = Sizeof(Key)) then
            begin
            Result := YUBIKEY_ERROR_INVALID_OTP;
            // User's UID and AesKey have been found. Now I need to decrypt the OTP
            YubikeyParse(Pointer(OTP), @Key[0], @Token);
            // First check: Checksum must match
            if YubikeyCRC16(@Token, Sizeof(Token)) = YUBIKEY_CRC_OK_RESIDUE then
              begin
              // Second check: UID must match
              b := True;
              for i := 0 to Pred(Sizeof(UID)) do
                if UID[i] <> Token.uid[i] then
                  begin
                  b := False;
                  Break;
                  end;
              if b then
                begin
                // Third check: Decrypted counter must be larger than stored counter
                RegCounterSize := Sizeof(RegCounter);
                if (RegQueryValueEx(Reg, YUBIKEY_REG_COUNTER, nil, nil, @RegCounter, @RegCounterSize) <> ERROR_SUCCESS) then
                  RegCounter := 0
                else if RegCounterSize <> Sizeof(RegCounter) then
                  RegCounter := 0;
                OtpCounter := Token.ctr and $7fff;
                OtpCounter := OtpCounter shl 16 + Token.use;
                if OtpCounter > RegCounter then
                  begin
                  // It's a valid OTP. Result code is set accordingly and the new Counter value is stored into registry
                  Result := 0;
                  RegSetValueEx(Reg, YUBIKEY_REG_COUNTER, 0, REG_DWORD, @OtpCounter, SizeOf(OtpCounter));
                  end;
                end;
              end
            end
          else
            Result := YUBIKEY_ERROR_INVALID_DATA
        else
          Result := YUBIKEY_ERROR_UNKNOWN_NAME;
        end;
    finally
      CloseRegistryKey(Mutex, OwnMutex, Reg);
      end;
    end;
end;

end.
